from __future__ import print_function
from __future__ import absolute_import
from uniRW.util import idenL

class State:

    def __init__(self, init_state, update_func=idenL):
        self.__state = init_state
        self.__update = update_func
        self.__read_only = True

    def check(self, name):
        return name in self.__state

    def get(self, name):
        if name not in self.__state:
            raise KeyError(str(name) + " is not in state")
        else:
            return self.__state[name]

    def set(self, name, val):
        if not self.__read_only:
            if name not in self.__state:
                raise KeyError(str(name) + " is not in state")
            else:
                self.__state[name] = val
        else:
            raise KeyError("Error: cannot change state")

    def set_update(self, update_func):
        self.__update = update_func

    def lock(self):
        self.__read_only = True

    def release(self):
        self.__read_only = False

    def update(self, val):
        if not self.__read_only:
            self.__update(self, val)
        else:
            raise KeyError("Error: cannot change state")

    def print(self):
        print(self.__state)



from re import split as resplit

class Line:

    def __init__(self, delimiter, name_col_dict={}):
        self.name_col_dict = name_col_dict.copy()
        self.delimiter = delimiter
        self.__line = None

    def get_by_name(self, name):
        if name in self.name_col_dict:
            return self.__line[self.name_col_dict[name]]
        else:
            raise KeyError(str(name) + " is not in a data line")

    def get_by_index(self, index):
        return self.__line[index]

    def set(self, line):
        self.__line = resplit(self.delimiter, line)

    def set_header(self):
        for i,val in enumerate(self.__line):
            if val in self.name_col_dict:
                continue
            else:
                self.name_col_dict[val] = i

    def clear(self):
        self.name_col_dict = {}
        self.__line = None

class OutputLine:

    def __init__(self, delimiter, end_char='\n'):
        self.delimiter = delimiter
        self.end_char = end_char

    def get_line(self, values):
        return self.delimiter.join(values) + self.end_char


# utility functions

def check_and_apply_f(f_dict, f, key, arg):
  return f_dict[key](arg) if key in f_dict else f(arg)

def check_and_apply_2(f_dict, key, arg1, arg2):
  return f_dict[key](arg1,arg2) if key in f_dict else arg2

def idenL(x,dummy):
  return x

def idenR(dummy,x):
  return x

def pureR(f):
  return lambda _,x: f(x)

def pureL(f):
  return lambda x,_: f(x)


from __future__ import absolute_import
from os.path import isfile
from uniRW.Line import Line, OutputLine
from copy import copy

class File:

    def __init__(self, file_name, line):
        self.file_name = file_name
        self.line = copy(line)


class DataFile(File):

    def __init__(self, file_name, line, header_lineno=-1):
        if not isinstance(line, Line):
            raise ValueError("line is not a Line object.")
        File.__init__(self, file_name, line)
        if not isfile(file_name):
            raise ValueError(str(file_name) +" is not a file or not found")
        self.header_lineno = header_lineno
        self.__init_line = copy(line)

    def reset(self):
        self.line = copy(self.__init_line)

class OutputFile(File):

    def __init__(self, file_name, line, header=[], foreword=[], epilogue=[]):
        if not isinstance(line, OutputLine):
            raise ValueError("line is not a Line object.")
        File.__init__(self, file_name, line)
        self.header = copy(header)
        self.foreword = copy(foreword)
        self.epilogue = copy(epilogue)


from __future__ import print_function
from uniRW.Value import GeneralValue

class HWriter:

    def __init__(self, hierarchy_spec, value_line):
        self.hierarchy_spec = hierarchy_spec
        self.value_line = value_line

    def flatten(self, layer, value_hierarchy, value_lines, value_line={}):

        for root, next in value_hierarchy.items():

            if type(layer) is list:
                current_value_line = value_line
                for value in layer:
                    if isinstance(value, GeneralValue):
                        current_value_line[value.name] = value_hierarchy[value.name]
                    else:
                        val, next_layer = value
                        self.flatten(next_layer, next, value_lines, current_value_line)

            elif type(layer) is dict:
                current_value_line = value_line.copy()
                value, next_layer = list(layer.items())[0]
                current_value_line[value.name] = root
                self.flatten(next_layer, next, value_lines, current_value_line)
                value_lines.append(current_value_line)

            else:
                current_value_line = value_line
                val, next_layer = layer
                self.flatten(next_layer, next, value_lines, current_value_line)


    def write(self, out_file, value_hierarchy, mode='w', sort_by=None, reverse=False):

        output = open(out_file.file_name, mode)
        for foreword_line in out_file.foreword:
            print(foreword_line, file=output)
        if out_file.header != []:
            header_line = out_file.line.get_line(out_file.header)
        else:
            header_line = out_file.line.get_line([value.name for value in self.value_line])
        print(header_line, file=output, end='')

        value_lines = []
        self.flatten(self.hierarchy_spec, value_hierarchy, value_lines)
        if sort_by == None:
            sorted_lines = value_lines
        else:
            sorted_lines = sorted(value_lines, key=lambda v_dict: v_dict[sort_by])

        for line in sorted_lines:
            values = []
            for value in self.value_line:
                if isinstance(value, GeneralValue):
                    values.append(value.to_string(line[value.name]))
                else:
                    raise ValueError("Value is not a value object")
            print(out_file.line.get_line(values), file=output, end='')

        for epilogue_line in out_file.epilogue:
            print(epilogue_line, file=output)

        output.close()



from __future__ import absolute_import

from uniRW.RW import read, readAll, write
from uniRW.Key import Key, StateKey, GeneralKey
from uniRW.Value import Value, StateValue, GeneralValue
from uniRW.State import State
from uniRW.Line import Line, OutputLine
from uniRW.File import DataFile, OutputFile, File
from uniRW.Reader import Reader
from uniRW.Writer import Writer
from uniRW.HReader import HReader
from uniRW.HWriter import HWriter
from uniRW.util import *




from distutils.core import setup

setup(
  name = 'uniRW',
  packages = ['uniRW'],
  version = '0.4.6',
  description = 'A universal reader and writer for stateful data file processing',
  author = 'Langxuan Su',
  author_email = 'lawliet@orion.codes',
  url = 'https://github.com/law-liet/uniRW',
  download_url = 'https://github.com/law-liet/uniRW/tarball/v0.4.6',
  keywords = ['universal', 'reader', 'writer',  'map', 'reduce'],
  classifiers = [
    "Programming Language :: Python",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Topic :: Text Processing",
    "Topic :: Utilities",
    "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)


from uniRW.Value import Value, StateValue, GeneralValue
from uniRW.State import State
from uniRW.File import DataFile
from copy import copy

class HReader:

    def __init__(self, hierarchy_spec, state=None, filter_f=lambda _, x: True):
        self.hierarchy_spec = hierarchy_spec
        self.state = state
        self.filter_f = filter_f
        self.__init_state = copy(self.state)


    def apply_post_map(self, layer, state, value_hierarchy):
        if type(layer) is list:

            for value in layer:
                if isinstance(value, GeneralValue):
                    val_name = value.name
                    val = value.post_map_f(state, value_hierarchy[val_name])
                    value_hierarchy[val_name] = val
                else:
                    val, next_layer = value
                    self.apply_post_map(next_layer,state,value_hierarchy[val])


        elif type(layer) is dict:
            value, next_layer = list(layer.items())[0]
            for val, _ in value_hierarchy.items():
                self.apply_post_map(next_layer, state, value_hierarchy[val])
        else:
            val, next_layer = layer
            self.apply_post_map(next_layer,state,value_hierarchy[val])



    def merge(self, layer, value_hierarchy1, value_hierarchy2, post=False, state=None):
        merged_hierarchy = value_hierarchy1.copy()

        if type(layer) is list:

            for value in layer:

                if isinstance(value, GeneralValue):
                    val1 = value_hierarchy1[value.name]
                    val2 = value_hierarchy2[value.name]
                    if post:
                        new_val1 = value.post_map_f(state, val1)
                        new_val2 = value.post_map_f(state, val2)
                        new_val = value.post_reduce_f(new_val1, new_val2)
                    else:
                        new_val = value.reduce_f(val1,val2)
                    merged_hierarchy[value.name] = new_val

                else:
                    val, next_layer = value
                    merged_hierarchy[val] = \
                        self.merge(next_layer, value_hierarchy1[val], value_hierarchy2[val], post, state)

        elif type(layer) is dict:
            value, next_layer = list(layer.items())[0]
            for val, next_val_layer2 in value_hierarchy2.items():
                if val in value_hierarchy1:
                    next_val_layer1 = value_hierarchy1[val]
                    merged_hierarchy[val] = self.merge(next_layer, next_val_layer1, next_val_layer2, post, state)
                else:
                    if post:
                        self.apply_post_map(next_layer, state, next_val_layer2)
                    merged_hierarchy[val] = next_val_layer2
        else:
            val, next_layer = layer
            merged_hierarchy[val] = self.merge(next_layer, value_hierarchy1[val], value_hierarchy2[val], post, state)

            #raise ValueError("Invalid hierarchy_spec structure")

        return merged_hierarchy


    def traverse(self, layer, line, current_state, value_hierarchy):

        if type(layer) is list:

            for value in layer:

                if isinstance(value, StateValue):
                    val_name = value.name
                    if current_state.check(val_name):
                        val = value.map_f(current_state, current_state.get(val_name))
                    else:
                        raise KeyError(str(value.name) + " is not in state")
                elif isinstance(value, Value):
                    val_name = value.name
                    val = value.map_f(current_state, line.get_by_name(val_name))
                else:
                    val, next_layer = value
                    value_hierarchy[val] = {}
                    self.traverse(next_layer, line, current_state, value_hierarchy[val])

                if val_name in value_hierarchy:
                    current_val = value_hierarchy[val_name]
                    reduced_val = value.reduce_f(current_val, val)
                    value_hierarchy[val_name] = reduced_val
                else:
                    value_hierarchy[val_name] = val

        elif type(layer) is dict:

            value, next_layer = list(layer.items())[0]

            if isinstance(value, StateValue):
                val_name = value.name
                if current_state.check(val_name):
                    val = value.map_f(current_state, current_state.get(val_name))
                else:
                    raise KeyError(str(value.name) + " is not in state")
            elif isinstance(value, Value):
                val_name = value.name
                val = value.map_f(current_state, line.get_by_name(val_name))
            else:
                val, next_layer = value
                value_hierarchy[val] = {}
                self.traverse(next_layer, line, current_state, value_hierarchy[val])

            if val in value_hierarchy:
                new_hierarchy = {}
                self.traverse(next_layer, line, current_state, new_hierarchy)
                current_hierarchy = value_hierarchy[val]
                merged_hierarchy = self.merge(next_layer, current_hierarchy, new_hierarchy)
                value_hierarchy[val] = merged_hierarchy

            else:
                value_hierarchy[val] = {}
                self.traverse(next_layer, line, current_state, value_hierarchy[val])

        else:
            val, next_layer = layer
            value_hierarchy[val] = {}
            self.traverse(next_layer, line, current_state, value_hierarchy[val])

            #raise ValueError("Invalid hierarchy_spec structure")


    def read(self, data_file, mode='r', apply_post_map=False, carry_state=False):

        if not isinstance(data_file, DataFile):
            raise ValueError("Data file is not a DataFile object.")

        lineno = 0
        value_hierarchy = {}
        value_hierarchy_list = []
        current_state = None

        if type(self.hierarchy_spec) is list:
            multi_hierarchy = True
        else:
            multi_hierarchy = False

        if self.state != None:
            if not isinstance(self.state, State):
                raise ValueError("State is a State object")

            if not carry_state:
                current_state = copy(self.state)
            else:
                current_state = self.state

        with open(data_file.file_name, mode) as file:

             for line in file:
                try:
                    data_file.line.set(line[:-1])

                    if data_file.header_lineno == lineno:
                        data_file.line.set_header()
                        lineno += 1
                        continue

                    if self.state != None:
                        current_state.release()
                        current_state.update(data_file.line)
                        current_state.lock()

                    if not self.filter_f(current_state, data_file.line):
                        lineno += 1
                        continue

                    if multi_hierarchy:
                        for i, hierarchy_spec in enumerate(self.hierarchy_spec):
                            if i >= len(value_hierarchy_list):
                                value_hierarchy_list.append(value_hierarchy.copy())
                            value_hierarchy_copy = value_hierarchy_list[i]
                            self.traverse(hierarchy_spec, data_file, current_state, value_hierarchy_copy)
                            value_hierarchy_list[i] = value_hierarchy_copy

                    else:
                        self.traverse(self.hierarchy_spec, data_file.line, current_state, value_hierarchy)

                except (KeyError, ValueError):
                    print("Error occurred when reading line" + str(lineno) + " of " + data_file.file_name)
                    raise

                lineno += 1



        if multi_hierarchy:

            if apply_post_map:
                for hierarchy_spec, value_hierarchy in zip(self.hierarchy_spec, value_hierarchy_list):
                    self.apply_post_map(hierarchy_spec, current_state, value_hierarchy)

            return value_hierarchy_list, current_state
        else:
            if apply_post_map:
                self.apply_post_map(self.hierarchy_spec, current_state, value_hierarchy)

            return value_hierarchy, current_state


    def readAll(self, data_files, mode='r', carry_state=False):

        if type(self.hierarchy_spec) is list:
            multi_hierarchy = True
        else:
            multi_hierarchy = False

        if multi_hierarchy:
            final_value_hierarchy_list = []

            for data_file in data_files:
                value_hierarchy_list, final_state = self.read(data_file, mode, carry_state)

                for i, value_hierarchy in enumerate(value_hierarchy_list):
                    if i >= len(final_value_hierarchy_list):
                        final_value_hierarchy_list.append({})
                    final_value_hierarchy = final_value_hierarchy_list[i]
                    final_value_hierarchy_list[i] = \
                        self.merge(self.hierarchy_spec[i], final_value_hierarchy, value_hierarchy, True, final_state)

            self.clear_state()
            return final_value_hierarchy_list

        else:
            final_value_hierarchy = {}

            for data_file in data_files:
                value_hierarchy, final_state = self.read(data_file, mode, carry_state)
                final_value_hierarchy = \
                    self.merge(self.hierarchy_spec, final_value_hierarchy, value_hierarchy, True, final_state)

            self.clear_state()
            return final_value_hierarchy

    def clear_state(self):
        self.state = copy(self.__init_state)




from __future__ import print_function
from __future__ import absolute_import
from uniRW.util import idenR

class GeneralValue:

    def __init__(self, name, to_string, map_f, reduce_f, post_map_f, post_reduce_f):
        self.name = name
        self.map_f = map_f
        self.reduce_f = reduce_f
        self.to_string = to_string
        self.post_map_f = post_map_f
        self.post_reduce_f = post_reduce_f


class Value(GeneralValue):

    def __init__(self, name, to_string=str, map_f=idenR,
                 reduce_f=idenR, post_map_f=idenR, post_reduce_f=idenR):
        GeneralValue.__init__(self, name, to_string, map_f, reduce_f, post_map_f, post_reduce_f)


class StateValue(GeneralValue):

    def __init__(self, name, to_string=str, map_f=idenR,
                 reduce_f=idenR, post_map_f=idenR, post_reduce_f=idenR):
        GeneralValue.__init__(self, name, to_string,  map_f, reduce_f, post_map_f, post_reduce_f)
